#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include "smo.h"

// predict and save results

BOOL svm_predict ( Data_List * testlist, smo_Settings * settings )
{	
	Data_List * trainlist ;
	Data_Node * trainnode ;
	Data_Node * testnode ;
	double fx, kernel, guess = 0 ;
	double error = 0 ;
	double alpha ;
	unsigned int i, j=0, k ;

	if (testlist == NULL || settings == NULL)
		return FALSE ;
	
	trainlist = settings->pairs ;

	if (TRUE == Is_Data_Empty(testlist))
		return FALSE ;	
	if (trainlist->dimen != testlist->dimen)
		return FALSE ;

	settings->c1p = 0 ;
	settings->c1n = 0 ;
	settings->c2p = 0 ;	
	settings->c2n = 0 ;		
	settings->svs = 0 ;

	i = 0 ;
	testnode = testlist->front ;
	while (testnode!=NULL)
	{ 
		fx = 0 ;
		// transform input point
		if (TRUE == trainlist->normalized_input)
		{	
			for (k=0;k<trainlist->dimen;k++)
			{ 
				if ( 0 != trainlist->x_devi[k] )
					testnode->point[k] = (testnode->point[k]-trainlist->x_mean[k])/(trainlist->x_devi[k]) ;
				else
					testnode->point[k] = 0 ;
			}
		}
		trainnode = trainlist->front ;
		j = 0 ;
		while (trainnode!=NULL)
		{		
			// calculate kernel 
			alpha = 0 ;
			for (k=0;k<settings->pairs->classes-1;k++)
			{
				if ((ALPHA+j)->pair->target<=k+1)
					alpha -= (ALPHA+j)->alpha[k] ;
				else
					alpha += (ALPHA+j)->alpha[k] ;
			}
			if ( alpha != 0 )
			{
				kernel = Calculate_Kernel (trainnode->point, testnode->point, settings) ;				
				fx = fx + alpha * kernel ;
				if (i==0)
					settings->svs ++ ;
			}
			trainnode = trainnode->next ;
			j++ ;
		}
		testnode -> fx = fx ;
		testnode->guess = fx ;
		
		if ( ORDINAL == trainlist->datatype )
		{
			// save guess of target in datafile	
			testnode->guess = 1 ;
			for (k=1;k<settings->pairs->classes;k++)
			{
				if (fx>settings->biasj[k-1])
					testnode->guess = k+1 ;
				else
					k = settings->pairs->classes ;
			}
			error += fabs(testnode->guess-testnode->target) ;
			if (fabs(testnode->guess-testnode->target)>0.5)
				guess += 1.0 ;
		}
		testnode = testnode->next ;
		i++ ;
	}
	if (i==testlist->count&&j==trainlist->count)
	{
		if ( ORDINAL == trainlist->datatype )
		{			
			settings->testrate = error/testlist->count ; //aae error 
			settings->testerror = guess/testlist->count ; //error number
		}
	}
	else
	{
		printf("Error in smo Prediction.\r\n") ;		
		settings->svs = 0 ;
		settings->testrate = 0 ;
		settings->testerror = 0 ;
	}	
	return TRUE ;
}

BOOL svm_saveresults ( Data_List * testlist, smo_Settings * settings )
{
	Data_Node * testnode ;
	unsigned int i, result, k ;
	char buf[1000] = "" ;
	char * pstr ;
	double temp ;
	double alpha ;
	FILE * guess_file = NULL ;
	FILE * svmfunc = NULL ;
	FILE * svmalpha = NULL ;
	FILE * svmresu = NULL ;	
	FILE * testfunc = NULL ;
	FILE * svmweight = NULL ;
	
	if (testlist == NULL || settings == NULL)
		return FALSE ;
	
	if (TRUE == Is_Data_Empty(testlist))
		return FALSE ;	
	if (settings->pairs->dimen != testlist->dimen)
		return FALSE ;

	strcpy (buf, INPUTFILE) ;
	strcat (buf,".svm.conf") ;
	svmfunc = fopen( buf, "w+t" ) ; 	
	printf("FUNCTION VALUES for training data have been saved in %s.\n",buf) ;
	
	strcpy (buf, INPUTFILE) ;
	strcat (buf,".svm.alpha") ;
	svmalpha = fopen( buf, "w+t" ) ;
	printf("ALPHAS have been saved in %s.\n",buf) ;
	
	strcpy (buf, INPUTFILE) ;
	strcat (buf,".svm.resu") ;
	svmresu = fopen( buf, "w+t" ) ;
	//printf("FUNCTION VALUES for training data have been saved in %s.\n",buf) ;

	if ((POLYNOMIAL==KERNEL && 1==P) || LINEAR==KERNEL )
	{
		strcpy (buf, INPUTFILE) ;
		strcat (buf,".svm.weight") ;
		svmweight = fopen( buf, "w+t" ) ;
		printf("LINEAR WEIGHTS have been saved in %s.\n",buf) ;
	}

	if (testlist != settings->pairs)
	{
	pstr = strstr( INPUTFILE, "train" ) ;
	if (NULL != pstr)
	{
		result = abs( INPUTFILE - pstr ) ;
		strncpy (buf, INPUTFILE, result ) ;
		buf[result] = '\0' ;
		strcat(buf, "cguess") ;
		strcat (buf, pstr+5) ;
	}
	else
	{
		strcpy (buf, INPUTFILE) ;
		strcat (buf,".test.resu") ;
	}
	guess_file = fopen( buf, "w+t" ) ;	
	printf("PREDICTIVE LABELS for test data have been saved in %s.\n",buf) ;	
	strcat(buf, ".svm.conf") ;
	testfunc = fopen( buf, "w+t" ) ; 	
	printf("FUNCTION VALUES for test data have been saved in %s.\n",buf) ;
	
	//save test results now
	i = 0 ;
	testnode = testlist->front ;
	while (testnode!=NULL)
	{ 
		if (NULL!=testfunc)
		{
			sprintf(buf,"%f\r", testnode->fx) ;
			fwrite (buf, sizeof(char), strchr(buf,'\0')-buf, testfunc ) ;
		}		
		if (NULL!=guess_file)
		{
			sprintf(buf,"%.0f\r", testnode->guess) ;
			fwrite (buf, sizeof(char), strchr(buf,'\0')-buf, guess_file ) ;
		}			
		testnode=testnode->next ;
		i++ ;
	}
	if (i!=testlist->count)
		printf("Error : in the data list for TESTING.\r\n") ;	
	}
				
	if ( NULL != svmalpha )	
	{
		for (i = 0; i < settings->pairs->count; i ++)
		{
			alpha = 0 ;
			for (k=0;k<settings->pairs->classes-1;k++)
			{
				if ((ALPHA+i)->pair->target<=k+1)
					alpha -= (ALPHA+i)->alpha[k] ;
				else
					alpha += (ALPHA+i)->alpha[k] ;
			}
			fprintf(svmalpha,"%.12f\r", alpha) ;
		}
		for (i = 1; i < settings->pairs->classes; i ++)
			fprintf(svmalpha,"%.12f\r", settings->biasj[i-1]) ;
	}

	if ( NULL != svmfunc )	
	{
		for (i = 0; i < settings->pairs->count; i ++)
		{			
			(settings->alpha + i)->pair->guess = (settings->alpha + i)->f_cache ;	
			fprintf(svmfunc,"%.12f\r", fabs((settings->alpha + i)->pair->guess)) ;
		}
	}

	temp = 0 ;
	if ( NULL != svmresu )	
	{
		for (i = 0; i < settings->pairs->count; i ++)
		{
			if (ORDINAL == settings->pairs->datatype)
			{
				(settings->alpha + i)->pair->guess = 1 ;
				for (k=1;k<settings->pairs->classes;k++)
				{
					if ((settings->alpha + i)->f_cache>settings->biasj[k-1])
						(settings->alpha + i)->pair->guess = k+1 ;
					else
						k = settings->pairs->classes ;
				}
				if ( fabs((settings->alpha + i)->pair->guess-(settings->alpha + i)->pair->target) >= 1.0 )
					temp += 1 ;
			}
			else
				printf("\nFATAL ERROR : the type is not ORDINAL REGRESSION.\n") ;
		}
		printf("\nTraining Error %.0f\n", temp) ;
	}

	if ((POLYNOMIAL==KERNEL && 1==P) || LINEAR==KERNEL )
	{
		for (result=0;result<settings->pairs->dimen;result++)
		{
			temp=0 ;
			for (i = 0; i < settings->pairs->count; i ++)
			{
				alpha = 0 ;
				for (k=0;k<settings->pairs->classes-1;k++)
				{
					if ((ALPHA+i)->pair->target<=k+1)
						alpha -= (ALPHA+i)->alpha[k] ;
					else
						alpha += (ALPHA+i)->alpha[k] ;
				}				
				temp += (settings->ard[result]) * alpha * (settings->alpha + i)->pair->point[result] ;
			}
			fprintf(svmweight,"%.12f\r", temp) ;
		}
	}
	if (NULL != guess_file)
		fclose(guess_file) ;
	if (NULL != svmresu)
		fclose(svmresu) ;
	if (NULL != svmalpha)
		fclose(svmalpha) ;
	if (NULL != svmfunc)
		fclose(svmfunc) ;
	if (NULL != testfunc)
		fclose(testfunc) ;
	if (NULL != svmweight)
		fclose(svmweight) ;
	return TRUE ;
}
// end of svc_predict.c
